/*

Licensed under the GNU AFFERO GENERAL PUBLIC LICENSE Version 3 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package application

import (
	settings "k8s.libre.sh/application/settings"
	"k8s.libre.sh/interfaces"
	"k8s.libre.sh/meta"
	"k8s.libre.sh/status"
)

type Application struct {
	meta.Instance
	Settings   map[string]settings.Component
	Components map[string]Component
	Jobs       map[string]Component
	Owner      interfaces.ObjectWithStatus
	Status     status.ApplicationStatus
	Version    string
}

func (app *Application) GetVersion() string {
	return app.Version
}

func (app *Application) GetJobs() map[string]Component {
	return app.Jobs
}

func (app *Application) GetApplicationStatus() status.ApplicationStatus {
	return app.Status
}

func (app *Application) GetComponents() map[string]Component {
	return app.Components
}

func (app *Application) GetOwner() interfaces.ObjectWithStatus {
	return app.Owner
}

func (app *Application) GetSettings() map[string]settings.Component {
	return app.Settings
}

func (app *Application) SetDefaults() {}

func NewApplication(i Instance) *Application {

	app := &Application{}
	app.Settings = make(map[string]settings.Component, len(i.GetSettings()))
	app.Components = make(map[string]Component, len(i.GetComponents()))
	app.Jobs = make(map[string]Component, len(i.GetJobs()))

	for k, v := range i.GetSettings() {
		v.SetDefaults()
		app.Settings[k] = settings.NewSettings(v)
	}

	for k, v := range i.GetComponents() {
		v.SetDefaults()
		app.Components[k] = v
	}

	for k, v := range i.GetJobs() {
		v.Init()
		v.SetDefaults()
		app.Jobs[k] = v
	}

	om := &meta.ObjectMeta{
		Labels:      i.GetLabels(),
		Annotations: i.GetLabels(),
		Name:        i.GetName(),
		Namespace:   i.GetNamespace(),
	}
	app.Instance = om
	app.Owner = i.GetOwner()
	app.Version = i.GetVersion()
	app.Status = i.GetApplicationStatus()
	return app
}
