/*

Licensed under the GNU AFFERO GENERAL PUBLIC LICENSE Version 3 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package application

import (
	"context"
	"fmt"

	"github.com/hashicorp/go-version"
	"github.com/presslabs/controller-util/syncer"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	settings "k8s.libre.sh/application/settings"
	"k8s.libre.sh/application/settings/parameters"
	"k8s.libre.sh/interfaces"
	"k8s.libre.sh/meta"
	objects "k8s.libre.sh/objects"
	"k8s.libre.sh/status"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/reconcile"
)

type Instance interface {
	meta.Instance
	GetComponents() map[string]Component
	GetJobs() map[string]Component
	//	SyncOrder() map[int]string
	GetOwner() interfaces.ObjectWithStatus
	GetSettings() map[string]settings.Component
	GetApplicationStatus() status.ApplicationStatus
	SetDefaults()
}

type InstanceReconciler struct {
	ReconcilerBase *ReconcilerBase
	Instance       Instance
	SyncOrder      map[int]string
	//	ctx            context.Context
}

func NewInstanceReconciler(r *ReconcilerBase, i Instance, o map[int]string /* , ctx context.Context */) *InstanceReconciler {
	return &InstanceReconciler{
		ReconcilerBase: r,
		Instance:       i,
		SyncOrder:      o,
		//		ctx:            ctx,
	}
}

// ReconcileStatusAware represnt a CRD type that has been enabled with ReconcileStatus, it can then benefit of a series of utility methods.
type Status interface {
	GetStatus() status.ApplicationStatus
	SetStatus(conditions status.ApplicationStatus)
}

func (r *InstanceReconciler) Reconcile(request reconcile.Request) (reconcile.Result, error) {

	/* 	ctx, cancel := context.WithTimeout(context.Background(), reconcileTimeout)
	   	defer cancel() */

	err := Init(r.Instance, r.ReconcilerBase)
	if err != nil {
		return reconcile.Result{}, err
	}

	// Syncers
	syncers := make(map[string]map[int]syncer.Interface)
	settingsSyncers := make(map[int]syncer.Interface)
	syncers["settings"] = settingsSyncers

	// Settings Object Syncers
	for _, s := range r.Instance.GetSettings() {
		//	for _, s := range sett {
		for k, obj := range settings.GetObjects(s) {
			settingsSyncers[k] = objects.NewObjectSyncer(obj, r.Instance.GetOwner(), r.ReconcilerBase)
		}
	}

	jobs := r.Instance.GetJobs()

	current, err := version.NewVersion(r.Instance.GetApplicationStatus().Version)
	// get nextcloud desired version in spec
	desired, err := version.NewVersion(r.Instance.GetVersion())

	if len(r.Instance.GetApplicationStatus().Version) == 0 {
		//	if current == nil {
		delete(jobs, "upgrade")
	} else {
		switch current.Compare(desired) {
		case 1:
			// current version is greater than desired, no downgrade
			// TODO Format error
			return ctrl.Result{}, err
		case -1:
			// current version lower than desired, upgrade job
			delete(jobs, "install")
		case 0:
			// current version and desired are the same, normal start
			delete(jobs, "install")
			delete(jobs, "upgrade")
		}
	}

	for _, j := range jobs {
		jobSyncers := NewObjectSyncersFromComponent(j, r.ReconcilerBase, r.Instance.GetOwner())
		for _, js := range jobSyncers {
			objStatus, err := SyncObject(context.TODO(), r.ReconcilerBase, js)
			if err != nil {
				return ctrl.Result{}, err
			}
			// TODO add objects status to appStatus
			fmt.Println(objStatus)
		}

	}

	// Components Object Syncers
	for _, c := range r.Instance.GetComponents() {
		syncers[c.GetComponent()] = NewObjectSyncersFromComponent(c, r.ReconcilerBase, r.Instance.GetOwner())
	}

	oldStatus := r.Instance.GetApplicationStatus()
	appStatus, err := Sync(context.TODO(), r.ReconcilerBase, syncers, r.SyncOrder)

	if err != nil {
		return ctrl.Result{}, err
	}

	if err != nil {
		return ctrl.Result{}, err
	}

	//	nc.Status.ApplicationStatus = appStatus

	// Reconcile status
	/* 	nc.Status.Version = nc.Spec.Version
	   	nc.Status.Phase = appsv1alpha1.PhaseRunning */

	if &oldStatus != &appStatus {
		diff := status.Difference(oldStatus.GetResources(), appStatus.GetResources())
		if len(diff.Objects) > 0 {
			for _, o := range diff.Objects {
				unstruct := &unstructured.Unstructured{}
				unstruct.SetGroupVersionKind(o.GVK())
				unstruct.SetName(o.Name)
				unstruct.SetNamespace(r.Instance.GetNamespace())
				// TODO user finalizers ?
				deleteBackground := metav1.DeletePropagationBackground
				opts := client.DeleteOptions{
					PropagationPolicy: &deleteBackground,
				}
				if err := r.ReconcilerBase.GetClient().Delete(context.TODO(), unstruct, &opts); err != nil {
					return ctrl.Result{}, err
				}
			}
		}

		appStatus.Version = desired.String()
		owner := r.Instance.GetOwner()
		owner.SetApplicationStatus(appStatus)

		if err := r.ReconcilerBase.GetClient().Status().Update(context.TODO(), owner); err != nil {
			return ctrl.Result{}, err
		}
	}

	return ctrl.Result{}, nil
}

// func GetTemplateValues(i Instance) map[string]interface{} {
func GetTemplateValues(i Instance, setts map[string]settings.Component) map[string]interface{} {
	values := make(map[string]interface{})
	cptsVals := make(map[string]interface{})
	settsVals := make(map[string]interface{})

	for _, cpt := range i.GetComponents() {
		// TO FIX
		cptsVals[cpt.GetComponent()] = cpt.(interface{})
	}

	for _, sett := range setts {
		// TODO create templateValues for Settings interface
		settsVals[sett.GetMeta().GetComponent()] = parameters.KeyPairValues(sett.GetConfig().GetParameters())
	}

	values["components"] = cptsVals
	values["settings"] = settsVals

	return values

}

func InitSettings(i Instance, r interfaces.Reconcile) error {

	setts := i.GetSettings()

	// First we init external values and random values
	for _, sett := range setts {
		err := settings.Init(sett, r.GetClient(), i.GetOwner())

		if err != nil {
			return err
		}
	}

	// Transform instance to template values used to init settings
	values := GetTemplateValues(i, i.GetSettings())

	// Init the values from templates
	for _, sett := range setts {
		err := sett.GetConfig().GetParameters().InitTemplateValues(values)
		if err != nil {
			return err
		}
	}

	return nil
}

// Initiatilize & Defaults
func Init(i Instance, r interfaces.Reconcile) error {

	i.SetDefaults()

	// Create and Init Settings
	// sett, err := application.CreateAndInitSettings(app, r)
	err := InitSettings(i, r)

	if err != nil {
		return err
	}

	// Init Components
	cpts := i.GetComponents()
	InitComponentsFromSettings(cpts, i.GetSettings())

	return nil
}
