package application

import (
	"context"
	"sort"
	"time"

	"github.com/presslabs/controller-util/syncer"
	batchv1 "k8s.io/api/batch/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"k8s.io/apimachinery/pkg/runtime"
	"k8s.io/apimachinery/pkg/util/wait"
	"k8s.io/client-go/rest"
	"k8s.io/client-go/tools/record"
	interfaces "k8s.libre.sh/interfaces"
	"k8s.libre.sh/status"
	kstatus "sigs.k8s.io/cli-utils/pkg/kstatus/status"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/controller/controllerutil"
	"sigs.k8s.io/controller-runtime/pkg/reconcile"
	logf "sigs.k8s.io/controller-runtime/pkg/runtime/log"
)

// ReconcilerBase is a base struct from which all reconcilers can be derived. By doing so your finalizers will also inherir a set of utility functions
// To inherit from reconciler just build your finalizer this way:
// type MyReconciler struct {
//   util.ReconcilerBase
//   ... other optional fields ...
// }
var log = logf.Log.WithName("reconciler")

type ReconcilerBase struct {
	// This client, initialized using mgr.Client() above, is a split client
	// that reads objects from the cache and writes to the apiserver
	client        client.Client
	scheme        *runtime.Scheme
	restConfig    *rest.Config
	recorder      record.EventRecorder
	ObjectSyncers map[int]syncer.ObjectSyncer
}

func NewReconcilerBase(client client.Client, scheme *runtime.Scheme, restConfig *rest.Config, recorder record.EventRecorder) ReconcilerBase {
	return ReconcilerBase{
		client:     client,
		scheme:     scheme,
		restConfig: restConfig,
		recorder:   recorder,
	}
}

//IsValid determines if a CR instance is valid. this implementation returns always true, should be overridden
func (r *ReconcilerBase) IsValid(obj metav1.Object) (bool, error) {
	return true, nil
}

//IsInitialized determines if a CR instance is initialized. this implementation returns always true, should be overridden
func (r *ReconcilerBase) IsInitialized(obj metav1.Object) bool {
	return true
}

// Reconcile is a stub function to have ReconsicerBase match the Reconciler interface. You must redefine this function
func (r *ReconcilerBase) Reconcile(request reconcile.Request) (reconcile.Result, error) {
	return reconcile.Result{}, nil
}

// GetClient returns the underlying client
func (r *ReconcilerBase) GetClient() client.Client {
	return r.client
}

//GetRestConfig returns the undelying rest config
func (r *ReconcilerBase) GetRestConfig() *rest.Config {
	return r.restConfig
}

// GetRecorder returns the underlying recorder
func (r *ReconcilerBase) GetRecorder() record.EventRecorder {
	return r.recorder
}

// GetScheme returns the scheme
func (r *ReconcilerBase) GetScheme() *runtime.Scheme {
	return r.scheme
}

func Record(owner runtime.Object, recorder record.EventRecorder, result syncer.SyncResult) error {
	if recorder != nil && owner != nil && result.EventType != "" && result.EventReason != "" && result.EventMessage != "" {
		if result.Operation != controllerutil.OperationResultNone {
			recorder.Eventf(owner, result.EventType, result.EventReason, result.EventMessage)
		}
	}

	return nil
}

func SyncObject(ctx context.Context, r interfaces.Reconcile, syncer syncer.Interface) (resource status.ObjectStatus, err error) {

	syncResult, err := syncer.Sync(ctx)
	if err != nil {
		Record(syncer.GetOwner(), r.GetRecorder(), syncResult)
		return resource, err
	}
	obj := syncer.GetObject().(interfaces.Object)

	err = WaitUntilCompletion(r, obj)
	if err != nil {
		Record(syncer.GetOwner(), r.GetRecorder(), syncResult)
		return resource, err
	}

	objStatus := ObjectStatus(obj)

	// TODO TOFIX
	objStatus.Status = "Ready"

	Record(syncer.GetOwner(), r.GetRecorder(), syncResult)

	return objStatus, nil
}

func SyncObjects(ctx context.Context, r interfaces.Reconcile, syncers map[int]syncer.Interface) (resources status.Resources, err error) {
	keys := []int{}

	for k := range syncers {
		keys = append(keys, k)
	}

	sort.Ints(keys)

	for _, k := range keys {

		objStatus, err := SyncObject(ctx, r, syncers[k])
		if err != nil {
			return resources, err
		}
		resources.Objects = append(resources.Objects, objStatus)

		/* 		syncResult, err := syncers[k].Sync(ctx)
		   		if err != nil {
		   			Record(syncers[k].GetOwner(), r.GetRecorder(), syncResult)
		   			return resources, err
		   		}
		   		obj := syncers[k].GetObject().(interfaces.Object)

		   		err = WaitUntilCompletion(r, obj)
		   		if err != nil {
		   			Record(syncers[k].GetOwner(), r.GetRecorder(), syncResult)
		   			return resources, err
		   		}

		   		objStatus := ObjectStatus(obj)

		   		// TODO TOFIX
		   		objStatus.Status = "Ready"

		   		Record(syncers[k].GetOwner(), r.GetRecorder(), syncResult)
		   		resources.Objects = append(resources.Objects, objStatus) */
	}

	return resources, nil
}

func ObjectStatus(obj interfaces.Object) (objStatus status.ObjectStatus) {
	objStatus.Link = obj.GetSelfLink()
	objStatus.Name = obj.GetName()
	objStatus.Kind = obj.GetObjectKind().GroupVersionKind().Kind
	objStatus.Group = obj.GetObjectKind().GroupVersionKind().Group
	objStatus.Version = obj.GetObjectKind().GroupVersionKind().Version

	return objStatus
}

func Sync(ctx context.Context, r interfaces.Reconcile, syncers map[string]map[int]syncer.Interface, order map[int]string) (appStatus status.ApplicationStatus, err error) {
	/* 	keys := []int{}
	   	appStatus.Components = make(map[string]*status.ComponentStatus, len(syncers))

	   	for k := range order {
	   		keys = append(keys, k)
	   	}

	   	sort.Ints(keys)

	   	for _, k := range keys {
	   		resources, err := SyncObjects(ctx, r, syncers[order[k]])
	   		appStatus.Components[order[k]] = &status.ComponentStatus{
	   			Resources: resources,
	   		}

	   		if err != nil {
	   			// manage error
	   			return appStatus, err
	   		}
	   	}
		   // manage success */

	keys := []int{}
	appStatus.Components = make(map[string]status.ComponentStatus)

	for k := range order {
		keys = append(keys, k)
	}

	sort.Ints(keys)

	for _, k := range keys {
		resources, err := SyncObjects(ctx, r, syncers[order[k]])
		appStatus.Components[order[k]] = status.ComponentStatus{
			Resources: resources,
		}
		/* 		appStatus.Components = append(appStatus.Components, status.ComponentStatus{
			Resources: resources,
		}) */

		if err != nil {
			// manage error
			return appStatus, err
		}
	}

	return appStatus, nil
}

func WaitUntilCompletion(r interfaces.Reconcile, obj runtime.Object) error {

	return wait.Poll(RetryInterval, ReadinessTimeout, func() (bool, error) {
		objectKey, _ := client.ObjectKeyFromObject(obj)

		if err := r.GetClient().Get(context.Background(), objectKey, obj); err != nil {
			return false, err
		}

		unstruct := &unstructured.Unstructured{}
		unstructuredObj, err := runtime.DefaultUnstructuredConverter.ToUnstructured(obj)

		if err != nil {
			return false, err
		}

		unstruct.Object = unstructuredObj

		s, err := kstatus.Compute(unstruct)

		// Check if job then check if completed
		obj, ok := obj.(*batchv1.Job)
		if ok {
			if obj.Status.Succeeded > 0 { // || job.Status.Failed > types.Int32(job.Spec.BackoffLimit)  {
				return true, nil
			}
			return false, nil

		}

		if err != nil {
			return false, err
		}
		if s.Status == kstatus.FailedStatus {
			return false, err
		}
		if s.Status == kstatus.UnknownStatus {
			return false, err
		}
		if s.Status == kstatus.InProgressStatus {
			return false, nil
		}

		return true, nil
	})
}

const (
	RetryInterval    = 50 * time.Millisecond
	RetryTimeout     = 2 * time.Second
	ReadinessTimeout = 10 * time.Minute
	GCTimeout        = 5 * time.Minute
)
