/*

Licensed under the GNU AFFERO GENERAL PUBLIC LICENSE Version 3 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package settings

import (
	"sort"
	"strings"

	"sigs.k8s.io/controller-runtime/pkg/client"

	"k8s.libre.sh/application/settings/parameters"
	"k8s.libre.sh/interfaces"
	"k8s.libre.sh/meta"
	"k8s.libre.sh/objects"
	"k8s.libre.sh/objects/configmap"
	"k8s.libre.sh/objects/secret"
)

const (
	// Generate settings as env variables in a configMap or Secret
	GenEnvFile SettingsGenerate = "envFile"
)

const (
	// Generate settings as env variables in a configMap or Secret
	SecretSettings SettingsType = "secret"
)

type SettingsGenerate string
type SettingsType string

// +kubebuilder:object:generate=true
type Settings struct {
	CreateOptions *CreateOptions `json:"createOptions,omitempty"`
	*SettingsSpec `json:",inline"`
}

// SettingsSpec defines a list of parameters and references to resources from which those parameters can be fetched
// Only secrets and configmaps in the same namespace are supported as references
// +kubebuilder:object:generate=true
/* type SettingsSpec struct {
	Sources []Source `json:"sources,omitempty"`
	// Parameters is a list of parameters
	*parameters.Parameters `json:"parameters,omitempty"`
} */

// +kubebuilder:object:generate=true
type CreateOptions struct {
	CommonMeta   *meta.ObjectMeta `json:"commonMeta,omitempty"`
	SecretMeta   *meta.ObjectMeta `json:"secretMeta,omitempty"`
	ConfigMeta   *meta.ObjectMeta `json:"configMeta,omitempty"`
	Generate     SettingsGenerate `json:"generate,omitempty"`
	SettingsType SettingsType     `json:"type,omitempty"`
}

func NewSettings(s Component) *Settings {
	return &Settings{
		CreateOptions: s.GetCreateOptions(),
		SettingsSpec: &SettingsSpec{
			Sources:    s.GetConfig().GetSources(),
			Parameters: s.GetConfig().GetParameters(),
		},
	}
}

func (opts *CreateOptions) Init() {
	if opts.CommonMeta == nil {
		opts.CommonMeta = &meta.ObjectMeta{}
	}

	if opts.CommonMeta.Labels == nil {
		opts.CommonMeta.Labels = make(map[string]string)
	}

	if opts.SecretMeta == nil {
		opts.SecretMeta = &meta.ObjectMeta{}
	}

	if opts.SecretMeta.Labels == nil {
		opts.SecretMeta.Labels = make(map[string]string)
	}

	if opts.ConfigMeta == nil {
		opts.ConfigMeta = &meta.ObjectMeta{}
	}

	if opts.ConfigMeta.Labels == nil {
		opts.SecretMeta.Labels = make(map[string]string)
	}

}

func (s *Settings) GetCreateOptions() *CreateOptions { return s.CreateOptions }

func (s *Settings) GetConfig() Config { return s.SettingsSpec }

func (s *Settings) GetMeta() meta.Instance { return s.CreateOptions.CommonMeta }

func (s *Settings) GetObjects() map[int]objects.Object {

	cm := &configmap.ConfigMap{
		ObjectMeta: s.CreateOptions.ConfigMeta,
	}

	secret := &secret.Secret{
		ObjectMeta: s.CreateOptions.SecretMeta,
	}

	objs := make(map[int]objects.Object, 2)

	genConfigParams := []string{}
	genSecretParams := []string{}

	for _, p := range *s.SettingsSpec.Parameters {
		// TODO TO FIX
		if (p.MountType == parameters.MountEnvFile || p.MountType == parameters.MountFile) &&
			// TODO TO FIX
			(p.Type == parameters.SecretParameter || p.Type == "") &&
			len(p.Value) > 0 {
			if p.Generate != parameters.GenerateTemplate && p.Generate != "" {
				genSecretParams = append(genSecretParams, p.Key)
			}
			secret.Parameters = append(secret.Parameters, p)
		}
		if (p.MountType == parameters.MountEnvFile || p.MountType == parameters.MountFile) &&
			p.Type == parameters.ConfigParameter &&
			len(p.Value) > 0 {
			if p.Generate != parameters.GenerateTemplate && p.Generate != "" {
				genConfigParams = append(genConfigParams, p.Key)
			}
			cm.Parameters = append(cm.Parameters, p)
		}
	}

	// Sort annotations as it is sorted in kubernetes and will have a diff otherwise
	sort.Strings(genConfigParams)
	sort.Strings(genSecretParams)

	// TODO TO FIX remove annotation if secret is not generated anymore
	if len(genSecretParams) > 0 {
		if len(secret.ObjectMeta.Annotations) == 0 {
			secret.ObjectMeta.Annotations = make(map[string]string, len(secret.ObjectMeta.Annotations))
		}
		secret.ObjectMeta.Annotations["settings.k8s.libre.sh/generate"] = strings.Join(genSecretParams, ",")
	}

	if len(genConfigParams) > 0 {
		if len(secret.ObjectMeta.Annotations) == 0 {
			cm.ObjectMeta.Annotations = make(map[string]string, len(genConfigParams))
		}
		secret.ObjectMeta.Annotations["settings.k8s.libre.sh/generate"] = strings.Join(genConfigParams, ",")
	}

	if len(secret.Parameters) > 0 {
		objs[0] = secret
	}

	if len(cm.Parameters) > 0 {
		objs[1] = cm
	}

	return objs
}

func (s *Settings) SetDefaults() {

}

// func (s *Settings) Init(r interfaces.Reconcile, owner interfaces.Object) error {
func (s *Settings) Init(c client.Client, owner interfaces.Object) error {

	if s.CreateOptions == nil {
		s.CreateOptions = &CreateOptions{}
	}
	s.CreateOptions.Init()

	meta.SetObjectMeta(s.CreateOptions.CommonMeta, s.CreateOptions.ConfigMeta)
	meta.SetObjectMeta(s.CreateOptions.CommonMeta, s.CreateOptions.SecretMeta)

	err := InitParametersValueFrom(s, c, owner)

	if err != nil {
		return err
	}

	s.InitRandValues()

	if err != nil {
		return err
	}

	return nil
}

// InitParametersValueFrom intialise the parameters with values provided in external resources in the same namespace
// All parameters values are filled from those resources
// Only Secrets and Configmaps are supported
//func InitParametersValueFrom(s *Settings, r interfaces.Reconcile, owner interfaces.Object) error {
// func InitParametersValueFrom(s *Settings, c client.Client, owner interfaces.Object) error {
/* 	err := InitParamsValueFrom(s, c, owner)
   	if err != nil {
   		return err
   	}
   	return nil */
/*
	params := parameters.Parameters{}
	cm := &corev1.ConfigMap{}
	sec := &corev1.Secret{}

	paramsBySecretSource, paramsByConfigMapSource := OrderByResourceRef(s.SettingsSpec)

	for k, v := range paramsBySecretSource {
		sec.SetNamespace(s.CreateOptions.CommonMeta.GetNamespace())
		sec.SetName(k)

		ps, err := parameters.MergeParametersFromResource(c, sec, owner, v)

		if err != nil {
			return err
		}

		params = append(params, ps...)

	}

	for k, v := range paramsByConfigMapSource {
		cm.SetName(k)
		cm.SetNamespace(s.CreateOptions.CommonMeta.GetNamespace())

		ps, err := parameters.MergeParametersFromResource(c, cm, owner, v)

		if err != nil {
			return err
		}

		params = append(params, ps...)

	}

	secretSrc := Source{
		Ref:  s.GetCreateOptions().SecretMeta.GetName(),
		Type: "secret",
	}
	configSrc := Source{
		Ref:  s.GetCreateOptions().ConfigMeta.GetName(),
		Type: "configmap",
	}

	// Reset sources, we do not want to mount orginal resources
	if s.CreateOptions.Generate == GenEnvFile {
		s.Sources = &Sources{}
	}

	s.Parameters = parameters.MergeParameters(s.Parameters, &params)

	for _, v := range *s.Parameters {
		if v.MountType == parameters.MountEnvFile || s.CreateOptions.Generate == GenEnvFile {
			v.MountType = parameters.MountEnvFile
			// add newly created resource ref to sources
			if v.Type == parameters.ConfigParameter {
				s.Sources = AppendSourceIfUnique(s.Sources, configSrc)
			} else if v.Type == parameters.SecretParameter || v.Type == "" {
				s.Sources = AppendSourceIfUnique(s.Sources, secretSrc)
			}
			// Reset valueFrom, we do not want to mount orginal resources
			v.ValueFrom = parameters.ValueFrom{}
		}

		if s.CreateOptions.SettingsType == "configmap" {
			v.Type = parameters.ConfigParameter
		} else if s.CreateOptions.SettingsType == "secret" {
			v.Type = parameters.SecretParameter
		}

		if v.MountType == parameters.MountFile && len(v.Ref) == 0 {
			if v.Type == parameters.ConfigParameter {
				v.Ref = s.GetCreateOptions().ConfigMeta.GetName()

			}
			if v.Type == parameters.SecretParameter {
				v.Ref = s.GetCreateOptions().SecretMeta.GetName()

			}
		}
	}

	return nil
*/
// }
