package parameters_test

import (
	"context"
	"fmt"
	"math/rand"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"

	. "k8s.libre.sh/application/settings/parameters"

	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

var _ = Describe("Utils", func() {

	Describe("Merge Parameters", func() {
		It("should merge two parameters into a new parameters", func() {

			srcParam := &Parameter{
				Key:   "PARAM",
				Value: "src_param",
			}

			extraSrcParam := &Parameter{
				Key:   "EXTRA_SRC_PARAM",
				Value: "extra_src_param",
			}

			destParam := &Parameter{
				Key:   "PARAM",
				Value: "dest_param",
			}

			extraDestParam := &Parameter{
				Key:   "EXTRA_DEST_PARAM",
				Value: "extra_dest_param",
			}

			srcParameters := &Parameters{
				srcParam,
				extraSrcParam,
			}
			destParameters := &Parameters{
				destParam,
				extraDestParam,
			}

			expectedObj := &Parameters{
				//	srcParam,
				extraSrcParam,
				destParam,
				extraDestParam,
			}

			//	parameters.InitValues()
			res := MergeParameters(srcParameters, destParameters)

			expectedObjByKey := ParametersByKey(expectedObj)
			resByKey := ParametersByKey(res)

			Expect(resByKey).To(Equal(expectedObjByKey))
			//	Expect(err).NotTo(HaveOccurred())
		})

		It("should merge parameters with data", func() {

			data := map[string]string{
				"FROM_PARAM":       "data_param",
				"EXTRA_DATA_PARAM": "extra_data_param",
			}

			dataParam1 := &Parameter{
				Key:   "FROM_PARAM",
				Value: "data_param",
			}

			dataParam2 := &Parameter{
				Key:   "EXTRA_DATA_PARAM",
				Value: "extra_data_param",
			}

			destParam := &Parameter{
				Key: "PARAM",
				ValueFrom: ValueFrom{
					FromKey: "FROM_PARAM",
				},
				Value: "dest_param",
			}

			// TODO Error if parameter is not in data
			// TODO defautl FromKey to Key
			extraDestParam1 := &Parameter{
				Key:   "EXTRA_DEST_PARAM",
				Value: "extra_dest_param",
				ValueFrom: ValueFrom{
					FromKey: "FROM_EXTRA_PARAM",
				},
			}

			destParameters := Parameters{
				destParam,
				extraDestParam1,
			}

			expectedParam := &Parameter{
				Key:       "PARAM",
				Value:     "data_param",
				ValueFrom: destParam.ValueFrom,
			}

			res, err := MergeParametersWithData(destParameters, data)
			By("returning error when parameter is not in data")
			Expect(err).To(HaveOccurred())

			By("merging parameters and data when some parameters are set")

			expectedObject := &Parameters{
				expectedParam,
			}

			newDestParameters := Parameters{
				destParam,
			}

			res, err = MergeParametersWithData(newDestParameters, data)

			Expect(err).NotTo(HaveOccurred())
			Expect(&res).To(Equal(expectedObject))

			By("returning parameters from data when parameters are empty")

			expectedObject = &Parameters{
				dataParam1,
				dataParam2,
			}

			emptyParams := Parameters{}

			res, err = MergeParametersWithData(emptyParams, data)

			Expect(err).NotTo(HaveOccurred())
			Expect(&res).To(Equal(expectedObject))
		})
	})

	Describe("Get Data From Resource", func() {
		It("should get data from resources", func() {

			secretName := fmt.Sprintf("secret-%d", rand.Int31())
			configName := fmt.Sprintf("config-%d", rand.Int31())

			sec := &corev1.Secret{
				ObjectMeta: metav1.ObjectMeta{
					Name:      secretName,
					Namespace: "default",
				},
				Data: map[string][]byte{
					"Bar": []byte("bar"),
				},
			}

			cm := &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Name:      configName,
					Namespace: "default",
				},
				Data: map[string]string{
					"Foo": "foo",
				},
			}

			err := c.Create(context.TODO(), sec)
			Expect(err).NotTo(HaveOccurred())

			err = c.Create(context.TODO(), cm)
			Expect(err).NotTo(HaveOccurred())

			owner := &corev1.Pod{}

			expectedCmData := map[string]string{
				"Foo": "foo",
			}

			expectedSecretData := map[string]string{
				"Bar": "bar",
			}

			obj := &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Name:      configName,
					Namespace: "default",
				},
			}
			res, err := GetDataFromResource(c, obj, owner)
			Expect(err).NotTo(HaveOccurred())
			Expect(res).To(Equal(expectedCmData))

			res, err = GetDataFromResource(c, sec, owner)
			Expect(err).NotTo(HaveOccurred())
			Expect(res).To(Equal(expectedSecretData))

		})
	})
})
