package settings_test

import (
	"context"
	"fmt"
	"math/rand"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"

	. "k8s.libre.sh/application/settings"
	"k8s.libre.sh/application/settings/parameters"
	"k8s.libre.sh/meta"
)

var _ = Describe("Settings", func() {
	Describe("InitParametersValueFrom", func() {
		var sec *corev1.Secret
		var cm *corev1.ConfigMap
		var secSource /* , cmSource */ Source
		var cpt *Settings
		var owner *corev1.Pod
		var commonMeta, configMeta, secretMeta *meta.ObjectMeta
		var parameterFromConfigMap *parameters.Parameter

		BeforeEach(func() {
			//	ns := "default"

			secretName := fmt.Sprintf("secret-%d", rand.Int31())
			configName := fmt.Sprintf("config-%d", rand.Int31())

			sec = &corev1.Secret{
				ObjectMeta: metav1.ObjectMeta{
					Name:      secretName,
					Namespace: "default",
				},
				Data: map[string][]byte{
					"Bar": []byte("bar"),
				},
			}

			cm = &corev1.ConfigMap{
				ObjectMeta: metav1.ObjectMeta{
					Name:      configName,
					Namespace: "default",
				},
				Data: map[string]string{
					"Foo": "foo",
				},
			}

			secSource = Source{
				Ref:  secretName,
				Type: "secret",
			}

			/* 			cmSource = Source{
				Ref:  configName,
				Type: "configmap",
			} */

			commonMeta = &meta.ObjectMeta{
				Name:      "test",
				Namespace: "default",
			}

			commonMeta.Labels = make(map[string]string)
			commonMeta.SetInstance("test")
			commonMeta.SetApplication("test")
			commonMeta.SetComponent("test")

			secretMeta = &meta.ObjectMeta{
				Name: "test",
			}
			configMeta = &meta.ObjectMeta{
				Name: "test",
			}

			parameterFromConfigMap = &parameters.Parameter{
				Key:  "FooFoo",
				Type: "configmap",
				ValueFrom: parameters.ValueFrom{
					Ref:     configName,
					FromKey: "Foo",
					RefType: "configmap",
				},
			}

			/* 			parameterFromSecret = &parameters.Parameter{
				Key:  "FooFoo",
				Type: "configmap",
				ValueFrom: parameters.ValueFrom{
					Ref:     configName,
					FromKey: "Foo",
					RefType: "configmap",
				},
			} */

			cpt = &Settings{
				CreateOptions: &CreateOptions{
					CommonMeta: commonMeta,
					SecretMeta: secretMeta,
					ConfigMeta: configMeta,
					// Generate:     GenEnvFile,
					SettingsType: "",
				},
				SettingsSpec: &SettingsSpec{
					Sources: &Sources{
						secSource,
						//	cmSource,
					},
					Parameters: &parameters.Parameters{
						parameterFromConfigMap,
					},
				},
			}

			owner = &corev1.Pod{}
			owner.SetName("test")
			owner.SetNamespace("default")

			err := c.Create(context.TODO(), sec)
			Expect(err).NotTo(HaveOccurred())

			err = c.Create(context.TODO(), cm)
			Expect(err).NotTo(HaveOccurred())
		})

		It("initialize parameters from valueFrom", func() {

			parameterFromConfigMap.Value = "foo"

			expectedParamFromSourceCm := &parameters.Parameter{
				Key:   "Bar",
				Value: "bar",
			}

			expectedParams := &parameters.Parameters{
				parameterFromConfigMap,
				expectedParamFromSourceCm,
			}

			err := InitParametersValueFrom(cpt, c, owner)

			Expect(err).NotTo(HaveOccurred())
			Expect(cpt.Parameters).To(Equal(expectedParams))

		})

		It("initialize parameters from valueFrom with generate envFile", func() {

			cpt.CreateOptions.Generate = GenEnvFile

			expectedParamFromSourceCm := &parameters.Parameter{
				Key:       "Bar",
				Value:     "bar",
				MountType: parameters.MountEnvFile,
			}

			err := InitParametersValueFrom(cpt, c, owner)

			parameterFromConfigMap.Value = "foo"
			parameterFromConfigMap.MountType = parameters.MountEnvFile
			expectedParams := &parameters.Parameters{
				parameterFromConfigMap,
				expectedParamFromSourceCm,
			}

			Expect(err).NotTo(HaveOccurred())
			Expect(cpt.Parameters).To(Equal(expectedParams))

		})

		It("initialize parameters from valueFrom with default settings to configmap", func() {

			cpt.CreateOptions.Generate = GenEnvFile
			cpt.CreateOptions.SettingsType = "configmap"

			expectedParamFromSourceCm := &parameters.Parameter{
				Key:       "Bar",
				Value:     "bar",
				MountType: parameters.MountEnvFile,
				Type:      parameters.ConfigParameter,
			}

			err := InitParametersValueFrom(cpt, c, owner)

			parameterFromConfigMap.Value = "foo"
			parameterFromConfigMap.MountType = parameters.MountEnvFile
			parameterFromConfigMap.Type = parameters.ConfigParameter
			expectedParams := &parameters.Parameters{
				parameterFromConfigMap,
				expectedParamFromSourceCm,
			}

			Expect(err).NotTo(HaveOccurred())
			Expect(cpt.Parameters).To(Equal(expectedParams))

		})

		It("initialize parameters from valueFrom with default settings to secret", func() {

			cpt.CreateOptions.Generate = GenEnvFile
			cpt.CreateOptions.SettingsType = "secret"

			expectedParamFromSourceCm := &parameters.Parameter{
				Key:       "Bar",
				Value:     "bar",
				MountType: parameters.MountEnvFile,
				Type:      parameters.SecretParameter,
			}

			err := InitParametersValueFrom(cpt, c, owner)

			parameterFromConfigMap.Value = "foo"
			parameterFromConfigMap.MountType = parameters.MountEnvFile
			parameterFromConfigMap.Type = parameters.SecretParameter

			expectedParams := &parameters.Parameters{
				parameterFromConfigMap,
				expectedParamFromSourceCm,
			}

			Expect(err).NotTo(HaveOccurred())
			Expect(cpt.Parameters).To(Equal(expectedParams))

		})
	})
})
