/*

Licensed under the GNU AFFERO GENERAL PUBLIC LICENSE Version 3 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package service

import (
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	interfaces "k8s.libre.sh/interfaces"
	meta "k8s.libre.sh/meta"

	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/util/intstr"
)

// SecretRef represents a reference to a Secret
type SecretRef string

// Host represents a valid hostname
type Host string

// +kubebuilder:object:generate=true
type Service struct {
	*meta.ObjectMeta `json:"meta,omitempty"`
	*ServiceSpec     `json:",inline"`
}

// +kubebuilder:object:generate=true
type ServiceSpec struct {
	Port Port `json:"port,omitempty"`
	// The list of ports that are exposed by this service.
	ExtraPorts []Port `json:"extraPorts,omitempty"`
	// type determines how the Service is exposed. Defaults to ClusterIP. Valid
	// options are ExternalName, ClusterIP, NodePort, and LoadBalancer.
	// "ExternalName" maps to the specified externalName.
	// "ClusterIP" allocates a cluster-internal IP address for load-balancing to
	// endpoints. Endpoints are determined by the selector or if that is not
	// specified, by manual construction of an Endpoints object. If clusterIP is
	// "None", no virtual IP is allocated and the endpoints are published as a
	// set of endpoints rather than a stable IP.
	// "NodePort" builds on ClusterIP and allocates a port on every node which
	// routes to the clusterIP.
	// "LoadBalancer" builds on NodePort and creates an
	// external load-balancer (if supported in the current cloud) which routes
	// to the clusterIP.
	// More info: https://kubernetes.io/docs/concepts/services-networking/service/#publishing-services-service-types
	// +optional
	Type corev1.ServiceType `json:"type,omitempty"`
}

// +kubebuilder:object:generate=true
type Port struct {
	Name     string          `json:"name,omitempty"`
	Port     int32           `json:"port,omitempty"`
	Protocol corev1.Protocol `json:"protocol,omitempty"`
}

func (obj *Service) Init() {
	if obj.ObjectMeta == nil {
		obj.ObjectMeta = &meta.ObjectMeta{}
	}

	if obj.ObjectMeta.Labels == nil {
		obj.ObjectMeta.Labels = make(map[string]string)
	}

	if obj.ServiceSpec == nil {
		obj.ServiceSpec = &ServiceSpec{}
	}
}

func (r *ServiceSpec) GetServiceType() corev1.ServiceType { return r.Type }

func (r *ServiceSpec) GetServicePorts() []corev1.ServicePort {

	ports := []corev1.ServicePort{}

	for _, p := range r.ExtraPorts {
		ports = append(ports, p.GetServicePort())
	}

	ports = append(ports, r.Port.GetServicePort())

	return ports
}

func (p *Port) GetServicePort() corev1.ServicePort {

	port := corev1.ServicePort{
		TargetPort: intstr.FromString(p.Name),
		Port:       p.Port,
		Name:       p.Name,
	}

	if len(p.Protocol) > 0 {
		port.Protocol = p.Protocol
	}

	return port
}

func (o *Service) Mutate(obj interfaces.Object) error {
	MutateService(o, obj.(*corev1.Service), o.ObjectMeta)
	meta.MutateMeta(o, obj)
	return nil
}

func (o *Service) GetObject() interfaces.Object {
	return &corev1.Service{
		TypeMeta: metav1.TypeMeta{
			Kind:       "Service",
			APIVersion: "v1",
		},
	}
}

func (o *Service) GetObjectKind() string  { return "Service" }
func (o *Service) GetObjectGroup() string { return "core" }
